// ----------------------------------------------------------------------------
// Main.java
// A HTML slide show generator.
// $Id: Main.java,v 1.6 2006/06/19 11:12:24 schreine Exp schreine $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package slideshow;

import java.io.*;
import java.util.*;

import org.w3c.dom.*;
import org.xml.sax.*;
import javax.xml.parsers.*;

public class Main
{
  // example configuration file:
  //
  // <slideshow>
  // <head>
  //   % declaration of default parameters
  //   % parameter #xyz# in template is replaced by reference to file adsf
  //   <param name="xyz" type="ref" value="adsf" />
  //   % parameter #xyz# in template is replaced by content of file adsf
  //   <param name="xyz" type="include" value="adsf" />
  //   % parameter #xyz# in template is replaced by "adsf"
  //   <param name="xyz" type="text" value="adsf" />
  //   % parameter #xyz# in template is replaced by reference to next slide
  //   <param name="xyz" type="next" />
  //   % parameter #xyz# in template is replaced by reference to previous slide
  //   <param name="xyz" type="prev" />
  //   % declaration of templates 
  //   % (potentially with default instantiations of parameters)
  //   <template name="abc" file="/asdf/adf" />
  //   <template name="abc" file="/asdf/adf">
  //       <param name="xyz" type="text" value="adsf" />
  //   </template>
  // </head>
  // <body>
  //   % sequence of slides 
  //   % each slide constructed from template by substituting parameters
  //   <slide template="abc">
  //     <param name="content" type="file" value="ad.txt">
  //     <param name="next" type="slide" value="thisslide">
  //   </slide>
  //   <slide template="abc" name="thisslide">
  //     <param name="content" type="file" value="ad.txt">
  //   </slide>
  // </body>
  // </slideshow>
  
  // version string
  private static final String VERSION = "0.2";
  
  // the name of the data directory
  private static final String DATANAME = "data";
  
  // the program arguments
  private static String configPath = null;  // name of configuration file
  private static String outputPath = null;  // path to output directory

  // the corresponding files
  private static File configFile = null;
  private static File outputDir = null;
  private static File dataDir = null; // subdir of outputDir with data files
  
  private static class Template
  {
    String name = null;              // template name
    String path = null;              // file path of template
    String value = null;             // file content of template
    HashMap params = null;           // parameter map (may be null)
    
    Template(String name, String path, String value, HashMap params)
    {
      this.name = name;
      this.path = path;
      this.value = value;
      this.params = params;
    }
  }
  
  private static class Parameter
  {
    String name = null;  // parameter name
    String type  = null; // type of parameter
    String value = null; // its value
    
    Parameter(String name, String type, String value)
    {
      this.name = name;
      this.type = type;
      this.value = value;
    }
  }
  
  // map template name to Template object
  private static HashMap templates = new HashMap();
  
  // map parameter name to Parameter object
  private static HashMap params = new HashMap();
  
  // map name of file copy in data directory to full path of original file
  private static HashMap data = new HashMap();
  
  // map slide name to Integer describing number of slide
  private static HashMap slides = new HashMap();
  
  /***************************************************************************
   * Generate slide show.
   * @param args list of arguments, see usageError()
   **************************************************************************/
  public static void main(String[] args)
  {
    System.out.println("RISC SlideShow " + VERSION + 
      " (c) Research Institute for Symbolic Computation");
    System.out.println(
      "Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>");
    int n = args.length;
    int i=0;
    while (i < n)
    {
      String arg = args[i];
      if (arg.equals("-i"))
      {
        if (configPath != null) usageError();
        i = i+1;
        if (i >= n) usageError();
        configPath = args[i];
      }
      else if (arg.equals("-o"))
      {
        if (outputPath != null) usageError();
        i = i+1;
        if (i >= n) usageError();
        outputPath = args[i];
      }
      else
        usageError();
      i = i+1;
    }
    if (configPath == null || outputPath == null) usageError();
    configFile = new File(configPath);
    outputDir = new File(outputPath);
    if (!configFile.canRead())
      error("cannot read file " + configFile.getAbsolutePath());
    makeDir(outputDir);
    dataDir = new File(outputDir, DATANAME);
    makeDir(dataDir);
    try
    {
      System.out.println
        ("----------------------------------------------------------------------");
      System.out.println("Configuration file: " + 
        configFile.getCanonicalPath());
      System.out.println("Output directory: " + 
        outputDir.getCanonicalPath());
      System.out.println
        ("----------------------------------------------------------------------");
    }
    catch(IOException e)
    {
      error("exception " + e.toString());
    }
    process();
  }
    
  /***************************************************************************
   * Print usage and exit.
   **************************************************************************/
  private static void usageError()
  {
    System.out.flush();
    System.err.println("Usage: slideshow.Main -i <configfile> -o <directory>");
    System.err.println("\t<configfile> path of configuration file");
    System.err.println("\t<directory> path of output directory");
    System.exit(-1);
  }
  
  /***************************************************************************
   * Print error message and exit
   **************************************************************************/
  private static void error(String message)
  {
    System.out.flush();
    System.err.println("Error: " + message);
    System.exit(-1);
  }
  
  /***************************************************************************
   * Make writeable directory (if it does not yet exist).
   * @param dir the directory path
   **************************************************************************/
  private static void makeDir(File dir)
  {
    if (dir.exists())
    {
      if (dir.isDirectory()) return;
      error("path " + dir.getAbsolutePath() + " does not denote a directory");
    }
    else
    {
      boolean okay = dir.mkdir();
      if (!okay) error("cannot create directory " + dir.getAbsolutePath());
    }
  }
  
  /***************************************************************************
   * Process configuration file and generate slide show in output directory.
   **************************************************************************/
  private static void process()
  {
    try
    {
      InputStream in = 
        new BufferedInputStream(new FileInputStream(configFile));
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(false);
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.parse(in);
      if (document == null) error("invalid configuration file " + configFile);
      Node root = document.getFirstChild();
      if (root == null) error("invalid configuration file " + configFile);
      removeWhitespace(root);
      process(root);
    }
    catch(ParserConfigurationException e)
    {
      error("exception " + e.toString());
    }
    catch(IOException e)
    {
      error("exception " + e.toString());
    }
    catch(SAXException e)
    {
      error("exception " + e.toString());
    }
  }
  
  /***************************************************************************
   * Removes all text nodes that only contain whitespace from DOM tree.
   * @param node the root of the tree that is to be cleaned from whitespace.
   **************************************************************************/
  public static void removeWhitespace(Node node)
  {
    Vector empty = new Vector();
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    for (int i=0; i<n; i++)
    {
      Node child = children.item(i);
      if (!(child instanceof Text))
      {
        removeWhitespace(child);
        continue;
      }
      String data = ((Text)child).getData();
      boolean iswhite = true;
      int l = data.length();
      for (int j=0; j<l; j++)
      {
        char ch = data.charAt(j);
        if (!Character.isWhitespace(ch))
        {
          iswhite = false;
          break;
        }
      }
      if (iswhite) empty.add(child);
    }
    Iterator iterator = empty.iterator();
    while (iterator.hasNext())
      node.removeChild((Node)iterator.next());
  }      
  
  /***************************************************************************
   * Process configuration file and generate slide show in output directory.
   * @param config root of configuration file.
   **************************************************************************/
  private static void process(Node config)
  {
    if (!config.getNodeName().equals("slideshow"))
      error("tag '" + config.getNodeName() + "' ('slideshow' expected)");
    NodeList children = config.getChildNodes();
    if (children.getLength() != 2)
      error("two children expected from element slideshow");
    Node head = children.item(0);
    Node body = children.item(1);
    if (!head.getNodeName().equals("head"))
      error("tag '" + config.getNodeName() + "' ('head' expected)");
    if (!body.getNodeName().equals("body"))
      error("tag '" + config.getNodeName() + "' ('body' expected)");
    System.out.println("Processing head...");
    processHead(head);
    System.out.println("Reading body...");
    readBody(body);
    System.out.println("Processing body...");
    processBody(body);
    System.out.println("Done.");
  }
  
  /****************************************************************************
   * Process head of configuration file.
   * @param head the head part of the file.
   **************************************************************************/
  private static void processHead(Node head)
  {
    NodeList children = head.getChildNodes();
    int n = children.getLength();
    for (int i=0; i<n; i++)
    {
      System.out.println("Processing element " + i + "...");
      Node child = children.item(i);
      String tag = child.getNodeName();
      if (tag.equals("param"))
      {
        Parameter param = processParameter(child);
        if (params.get(param.name) != null)
          error("parameter '" + param.name + "' declared twice in head");
        params.put(param.name, param);
        continue;
      }
      if (tag.equals("template"))
      {
        Template template = processTemplate(child);
        if (templates.get(template.name) != null)
          error("template '" + template.name + "' declared twice in head");
        templates.put(template.name, template);
        continue;
      }
      error("tag '" + tag + "' in head ('param' or 'template' expected)"); 
    }
  }
  
  /***************************************************************************
   * Process parameter node and return parameter info.
   * @param node the parameter node.
   * @return the node information.
   **************************************************************************/
  private static Parameter processParameter(Node node)
  {
    NamedNodeMap attributes = node.getAttributes();
    if (attributes == null) 
      error("no attributes in element 'param'");
    Attr nameAttr = (Attr)attributes.getNamedItem("name");
    if (nameAttr == null)
      error("no attribute 'name' in element 'param'");
    String name = nameAttr.getValue();
    Attr typeAttr = (Attr)attributes.getNamedItem("type");
    if (typeAttr == null)
      error("no attribute 'type' in element 'param' with name '" + name + "'");
    String type = typeAttr.getValue();
    if (!type.equals("ref") && !type.equals("include") && 
        !type.equals("text") && !type.equals("prev") &&
        !type.equals("next") && !type.equals("slide"))
      error("invalid type attribute '" + type + 
          "' in element 'param' with name '" + name + "'");     
    String value = null;
    if (!type.equals("next") && !type.equals("prev"))
    {
      Attr valueAttr = (Attr)attributes.getNamedItem("value");
      if (valueAttr == null)
        error("no attribute 'value' in element 'param' with name '" +
            name + "'");
      value = valueAttr.getValue();
    }
    return new Parameter(name, type, value);
  }
  
  /***************************************************************************
   * Process template node and return parameter info.
   * @param node the template node.
   * @return the node information.
   **************************************************************************/
  private static Template processTemplate(Node node)
  {
    NamedNodeMap attributes = node.getAttributes();
    if (attributes == null) 
      error("no attributes in element 'template'");
    Attr nameAttr = (Attr)attributes.getNamedItem("name");
    if (nameAttr == null)
      error("no attribute 'name' in element 'template'");
    String name = nameAttr.getValue();
    Attr fileAttr = (Attr)attributes.getNamedItem("file");
    if (fileAttr == null)
      error("no attribute 'file' in element 'template' with name '" + name +
        "'");
    String file = fileAttr.getValue();
    String tstring = read(file);
    HashMap tparams = getParameters(node);
    return new Template(name, file, tstring, tparams);
  }
  
  /***************************************************************************
   * Process parameter children of current node.
   * @param node a node whose only children are parameter nodes.
   * @return the parameter information (null, if no children)
   **************************************************************************/
  private static HashMap getParameters(Node node)
  {
    NodeList children = node.getChildNodes();
    if (children == null) return null;
    HashMap params = new HashMap();
    int n = children.getLength();
    for (int i=0; i<n; i++)
    {
      Parameter param = processParameter(children.item(i));
      if (params.get(param.name) != null)
        error("parameter '" + param.name + "' declared twice");
      params.put(param.name, param);
    }
    return params;
  }
  
 /****************************************************************************
  * Read body of configuration file to determine value of "slide" references.
  * @param body the body part of the file.
  ***************************************************************************/
 private static void readBody(Node body)
 {
   NodeList children = body.getChildNodes();
   int n = children.getLength();
   for (int i=0; i<n; i++)
   {
     System.out.println("Reading element " + i + "...");
     Node child = children.item(i);
     String tag = child.getNodeName();
     if (!tag.equals("slide"))
       error("tag '" + tag + "' in body ('slide' expected)");
     NamedNodeMap attributes = child.getAttributes(); 
     if (attributes == null) continue;
     Attr nameAttr = (Attr)attributes.getNamedItem("name");
     if (nameAttr == null) continue;
     String name = nameAttr.getValue();
     if (slides.containsKey(name))
       error("name '" + name + "' already used for another slide");
     slides.put(name, new Integer(i));
   }
 }
  
  /****************************************************************************
   * Process body of configuration file.
   * @param body the body part of the file.
   **************************************************************************/
  private static void processBody(Node body)
  {
    NodeList children = body.getChildNodes();
    int n = children.getLength();
    for (int i=0; i<n; i++)
    {
      System.out.println("Processing element " + i + "...");
      Node child = children.item(i);
      String tag = child.getNodeName();
      if (!tag.equals("slide"))
        error("tag '" + tag + "' in body ('slide' expected)");
      NamedNodeMap attributes = child.getAttributes(); 
      if (attributes == null) 
        error("no attributes in element 'slide'");
      Attr templateAttr = (Attr)attributes.getNamedItem("template");
      if (templateAttr == null)
        error("no attribute 'template' in element 'slide'");
      String template = templateAttr.getValue();
      HashMap sparams = getParameters(child);
      generateSlide(template, sparams, i, n); 
    }
  }
  
  /***************************************************************************
   * Get name of i-th slide.
   * @param i the index of the slide, slide indices start with 0.
   * @return the file name of the slide.
   **************************************************************************/
  private static String getFileName(int i)
  {
    if (i == 0)
      return "index.html";
    else
      return "index_" + i + ".html";
  }
  
  /***************************************************************************
   * Generate slide from template.
   * @param tname the name of the template.
   * @param sparams the slide parameters.
   * @param i index of slide
   * @param n total number of slides
   *************************************************************************/
  private static void generateSlide(String tname, HashMap sparams,
    int i, int n)
  {
    Template template = (Template)templates.get(tname);
    if (template == null)
      error("no template with name '" + tname + "' declared in head");
    String value = template.value;
    value = expand(value, sparams, i, n);
    value = expand(value, template.params, i, n);
    value = expand(value, params, i, n);
    if (value.indexOf('#') != -1)
    {
      // expand another time to allow parameters in parameter text files
      value = expand(value, sparams, i, n);
      value = expand(value, template.params, i, n);
      value = expand(value, params, i, n);
    }
    File file = new File(outputDir, getFileName(i));
    try
    {
      Writer writer = new FileWriter(file);
      writer.write(value);
      writer.close();
    }
    catch(IOException e)
    {
      error("Cannot write file " + file.toString() 
          + "(" + e.toString() + ")");
    }
  }
  
  /**************************************************************************
   * Return version of template expanded according to parameters.
   * @param value the template value
   * @param params the expansion parameters
   * @param i index of slide for expansion
   * @param n total number of slides expanded.
   * @return the expanded string
   *************************************************************************/
  private static String expand(String value, HashMap params, int i, int n)
  {
    String result = value;
    Iterator iterator = params.values().iterator();
    while (iterator.hasNext())
    {
      Parameter p = (Parameter)iterator.next();
      result = expand(result, p, i, n);
    }
    return result;
  }
  
  /**************************************************************************
   * Return version of template expanded according to parameter.
   * @param value the template value
   * @param param the expansion parameter
   * @param i index of slide for expansion
   * @param n total number of slides expanded.
   * @return the expanded string
   *************************************************************************/
  private static String expand(String template, Parameter param, int i, int n)
  {
    String key = "#" + param.name + "#";
    String value = null;
    if (param.type.equals("ref"))
    {
      value = DATANAME + File.separatorChar + getDataName(param.value);
    }
    else if (param.type.equals("include"))
    {
      value = read(param.value);
    }
    else if (param.type.equals("text"))
    {
      value = param.value;
    }
    else if (param.type.equals("prev"))
    {
      int index = i-1;
      if (index < 0) index = 0;
      value = getFileName(index);
    }
    else if (param.type.equals("next"))
    {
      int index = i+1;
      if (index > n-1) index = n-1;
      value = getFileName(index);
    }
    else if (param.type.equals("slide"))
    {
      Integer p = (Integer)slides.get(param.value);
      if (p == null) error("there is no slide named " + param.value);
      value = getFileName(p.intValue());
    }
    return template.replaceAll(key, value);
  }
  
  /***************************************************************************
   * Return content of text file as string.
   * @param file the name of a text file.
   * @return its content as a string.
   **************************************************************************/
  private static String read(String file)
  {
    try
    {
      FileReader reader = new FileReader(file);
      StringBuffer buffer = new StringBuffer();
      char[] b = new char[10000];
      while (true)
      {
        int m = reader.read(b);
        if (m == -1) break;
        buffer.append(b, 0, m);
      }
      return buffer.toString();
    }
    catch (FileNotFoundException e)
    {
      error("cannot read template file " + file
          + " (" + e.getMessage() + ")");
    }
    catch (IOException e)
    {
      error("errror in reading template file " + file
          + " (" + e.getMessage() + ")");
    }
    return null;
  }
  
  /***************************************************************************
   * Get name of local file copy.
   * @param path path to external file.
   * @return name of local copy of file in data directory.
   **************************************************************************/
  private static String getDataName(String path)
  {
    File file = new File(path);
    String name = file.getName();
    String apath = file.getAbsolutePath();
    String bpath = (String)data.get(name);
    if (bpath != null)
    {
      if (bpath.equals(apath)) return name;
      String base = null;
      String ext = null;
      int pos = name.lastIndexOf('.');
      if (pos == -1)
        base = name;
      else
      {
        base = name.substring(0, pos);
        ext = name.substring(pos+1);
      }
      int i = 0;
      do
      {
        name = base + "_" + i;
        if (ext != null) name = name + '.' + ext;
        bpath = (String)data.get(name);
        if (bpath != null && bpath.equals(apath)) return name;
        i = i+1;
      }
      while (bpath != null);
    }
    data.put(name, apath);
    copy(file, new File(dataDir, name));
    return name;
  }
  
  /***************************************************************************
   * Copy file from one location to another.
   * @param in the input file.
   * @param out the output file.
   **************************************************************************/
  private static void copy(File in, File out)
  {
    try
    {
      FileInputStream is = new FileInputStream(in);
      FileOutputStream os = new FileOutputStream(out);
      byte[] buffer = new byte[10000];
      while (true)
      {
        int n = is.read(buffer);
        if (n == -1) break;
        os.write(buffer, 0, n);
      }
      is.close();
      os.close();
    }
    catch(IOException e)
    {
      error("cannot copy file " + in.getAbsolutePath() + " to " +
          out.getAbsolutePath() + " (" + e.toString() + ")");
    }
  }
}
